// Copyright 2010 Dolphin Emulator Project
// Licensed under GPLv2+
// Refer to the license.txt file included.

#include "Core/DSP/Jit/DSPEmitter.h"

#include <algorithm>
#include <cstring>

#include "Common/Assert.h"
#include "Common/BitSet.h"
#include "Common/CommonTypes.h"
#include "Common/Logging/Log.h"

#include "Core/DSP/DSPAnalyzer.h"
#include "Core/DSP/DSPCore.h"
#include "Core/DSP/DSPHost.h"
#include "Core/DSP/DSPMemoryMap.h"
#include "Core/DSP/DSPTables.h"

using namespace Gen;

namespace DSP
{
namespace JIT
{
namespace x86
{
constexpr size_t COMPILED_CODE_SIZE = 2097152;
constexpr size_t MAX_BLOCK_SIZE = 250;
constexpr u16 DSP_IDLE_SKIP_CYCLES = 0x1000;

DSPEmitter::DSPEmitter()
    : blockLinks(MAX_BLOCKS), blockSize(MAX_BLOCKS), blocks(MAX_BLOCKS),
      compileSR{SR_INT_ENABLE | SR_EXT_INT_ENABLE}
{
  AllocCodeSpace(COMPILED_CODE_SIZE);

  CompileDispatcher();
  stubEntryPoint = CompileStub();

  // Clear all of the block references
  std::fill(blocks.begin(), blocks.end(), (DSPCompiledCode)stubEntryPoint);
}

DSPEmitter::~DSPEmitter()
{
  FreeCodeSpace();
}

void DSPEmitter::ClearIRAM()
{
  for (int i = 0x0000; i < 0x1000; i++)
  {
    blocks[i] = (DSPCompiledCode)stubEntryPoint;
    blockLinks[i] = nullptr;
    blockSize[i] = 0;
    unresolvedJumps[i].clear();
  }
  g_dsp.reset_dspjit_codespace = true;
}

void DSPEmitter::ClearIRAMandDSPJITCodespaceReset()
{
  ClearCodeSpace();
  CompileDispatcher();
  stubEntryPoint = CompileStub();

  for (int i = 0x0000; i < 0x10000; i++)
  {
    blocks[i] = (DSPCompiledCode)stubEntryPoint;
    blockLinks[i] = nullptr;
    blockSize[i] = 0;
    unresolvedJumps[i].clear();
  }
  g_dsp.reset_dspjit_codespace = false;
}

// Must go out of block if exception is detected
void DSPEmitter::checkExceptions(u32 retval)
{
  // Check for interrupts and exceptions
  TEST(8, M(&g_dsp.exceptions), Imm8(0xff));
  FixupBranch skipCheck = J_CC(CC_Z, true);

  MOV(16, M(&(g_dsp.pc)), Imm16(compilePC));

  DSPJitRegCache c(gpr);
  gpr.SaveRegs();
  ABI_CallFunction(DSPCore_CheckExceptions);
  MOV(32, R(EAX), Imm32(retval));
  JMP(returnDispatcher, true);
  gpr.LoadRegs(false);
  gpr.FlushRegs(c, false);

  SetJumpTarget(skipCheck);
}

bool DSPEmitter::FlagsNeeded() const
{
  const u8 flags = Analyzer::GetCodeFlags(compilePC);

  return !(flags & Analyzer::CODE_START_OF_INST) || (flags & Analyzer::CODE_UPDATE_SR);
}

void DSPEmitter::FallBackToInterpreter(UDSPInstruction inst)
{
  const DSPOPCTemplate* const op_template = GetOpTemplate(inst);

  if (op_template->reads_pc)
  {
    // Increment PC - we shouldn't need to do this for every instruction. only for branches and end
    // of block.
    // Fallbacks to interpreter need this for fetching immediate values

    MOV(16, M(&(g_dsp.pc)), Imm16(compilePC + 1));
  }

  // Fall back to interpreter
  gpr.PushRegs();
  _assert_msg_(DSPLLE, op_template->intFunc, "No function for %04x", inst);
  ABI_CallFunctionC16(op_template->intFunc, inst);
  gpr.PopRegs();
}

void DSPEmitter::EmitInstruction(UDSPInstruction inst)
{
  const DSPOPCTemplate* const op_template = GetOpTemplate(inst);
  bool ext_is_jit = false;

  // Call extended
  if (op_template->extended)
  {
    const DSPOPCTemplate* const ext_op_template = GetExtOpTemplate(inst);

    if (!ext_op_template->jitFunc)
    {
      // Fall back to interpreter
      gpr.PushRegs();
      ABI_CallFunctionC16(ext_op_template->intFunc, inst);
      gpr.PopRegs();
      INFO_LOG(DSPLLE, "Instruction not JITed(ext part): %04x", inst);
      ext_is_jit = false;
    }
    else
    {
      (this->*ext_op_template->jitFunc)(inst);
      ext_is_jit = true;
    }
  }

  // Main instruction
  if (!op_template->jitFunc)
  {
    FallBackToInterpreter(inst);
    INFO_LOG(DSPLLE, "Instruction not JITed(main part): %04x", inst);
  }
  else
  {
    (this->*op_template->jitFunc)(inst);
  }

  // Backlog
  if (op_template->extended)
  {
    if (!ext_is_jit)
    {
      // need to call the online cleanup function because
      // the writeBackLog gets populated at runtime
      gpr.PushRegs();
      ABI_CallFunction(applyWriteBackLog);
      gpr.PopRegs();
    }
    else
    {
      popExtValueToReg();
    }
  }
}

void DSPEmitter::Compile(u16 start_addr)
{
  // Remember the current block address for later
  startAddr = start_addr;
  unresolvedJumps[start_addr].clear();

  const u8* entryPoint = AlignCode16();

  /*
  // Check for other exceptions
  if (dsp_SR_is_flag_set(SR_INT_ENABLE))
    return;

  if (g_dsp.exceptions == 0)
    return;
  */

  gpr.LoadRegs();

  blockLinkEntry = GetCodePtr();

  compilePC = start_addr;
  bool fixup_pc = false;
  blockSize[start_addr] = 0;

  while (compilePC < start_addr + MAX_BLOCK_SIZE)
  {
    if (Analyzer::GetCodeFlags(compilePC) & Analyzer::CODE_CHECK_INT)
      checkExceptions(blockSize[start_addr]);

    UDSPInstruction inst = dsp_imem_read(compilePC);
    const DSPOPCTemplate* opcode = GetOpTemplate(inst);

    EmitInstruction(inst);

    blockSize[start_addr]++;
    compilePC += opcode->size;

    // If the block was trying to link into itself, remove the link
    unresolvedJumps[start_addr].remove(compilePC);

    fixup_pc = true;

    // Handle loop condition, only if current instruction was flagged as a loop destination
    // by the analyzer.
    if (Analyzer::GetCodeFlags(static_cast<u16>(compilePC - 1u)) & Analyzer::CODE_LOOP_END)
    {
      MOVZX(32, 16, EAX, M(&(g_dsp.r.st[2])));
      TEST(32, R(EAX), R(EAX));
      FixupBranch rLoopAddressExit = J_CC(CC_LE, true);

      MOVZX(32, 16, EAX, M(&g_dsp.r.st[3]));
      TEST(32, R(EAX), R(EAX));
      FixupBranch rLoopCounterExit = J_CC(CC_LE, true);

      if (!opcode->branch)
      {
        // branch insns update the g_dsp.pc
        MOV(16, M(&(g_dsp.pc)), Imm16(compilePC));
      }

      // These functions branch and therefore only need to be called in the
      // end of each block and in this order
      DSPJitRegCache c(gpr);
      HandleLoop();
      gpr.SaveRegs();
      if (!Host::OnThread() && Analyzer::GetCodeFlags(start_addr) & Analyzer::CODE_IDLE_SKIP)
      {
        MOV(16, R(EAX), Imm16(DSP_IDLE_SKIP_CYCLES));
      }
      else
      {
        MOV(16, R(EAX), Imm16(blockSize[start_addr]));
      }
      JMP(returnDispatcher, true);
      gpr.LoadRegs(false);
      gpr.FlushRegs(c, false);

      SetJumpTarget(rLoopAddressExit);
      SetJumpTarget(rLoopCounterExit);
    }

    if (opcode->branch)
    {
      // don't update g_dsp.pc -- the branch insn already did
      fixup_pc = false;
      if (opcode->uncond_branch)
      {
        break;
      }
      else if (!opcode->jitFunc)
      {
        // look at g_dsp.pc if we actually branched
        MOV(16, R(AX), M(&g_dsp.pc));
        CMP(16, R(AX), Imm16(compilePC));
        FixupBranch rNoBranch = J_CC(CC_Z, true);

        DSPJitRegCache c(gpr);
        // don't update g_dsp.pc -- the branch insn already did
        gpr.SaveRegs();
        if (!Host::OnThread() && Analyzer::GetCodeFlags(start_addr) & Analyzer::CODE_IDLE_SKIP)
        {
          MOV(16, R(EAX), Imm16(DSP_IDLE_SKIP_CYCLES));
        }
        else
        {
          MOV(16, R(EAX), Imm16(blockSize[start_addr]));
        }
        JMP(returnDispatcher, true);
        gpr.LoadRegs(false);
        gpr.FlushRegs(c, false);

        SetJumpTarget(rNoBranch);
      }
    }

    // End the block if we're before an idle skip address
    if (Analyzer::GetCodeFlags(compilePC) & Analyzer::CODE_IDLE_SKIP)
    {
      break;
    }
  }

  if (fixup_pc)
  {
    MOV(16, M(&(g_dsp.pc)), Imm16(compilePC));
  }

  blocks[start_addr] = (DSPCompiledCode)entryPoint;

  // Mark this block as a linkable destination if it does not contain
  // any unresolved CALL's
  if (unresolvedJumps[start_addr].empty())
  {
    blockLinks[start_addr] = blockLinkEntry;

    for (u16 i = 0x0000; i < 0xffff; ++i)
    {
      if (!unresolvedJumps[i].empty())
      {
        // Check if there were any blocks waiting for this block to be linkable
        size_t size = unresolvedJumps[i].size();
        unresolvedJumps[i].remove(start_addr);
        if (unresolvedJumps[i].size() < size)
        {
          // Mark the block to be recompiled again
          blocks[i] = (DSPCompiledCode)stubEntryPoint;
          blockLinks[i] = nullptr;
          blockSize[i] = 0;
        }
      }
    }
  }

  if (blockSize[start_addr] == 0)
  {
    // just a safeguard, should never happen anymore.
    // if it does we might get stuck over in RunForCycles.
    ERROR_LOG(DSPLLE, "Block at 0x%04x has zero size", start_addr);
    blockSize[start_addr] = 1;
  }

  gpr.SaveRegs();
  if (!Host::OnThread() && Analyzer::GetCodeFlags(start_addr) & Analyzer::CODE_IDLE_SKIP)
  {
    MOV(16, R(EAX), Imm16(DSP_IDLE_SKIP_CYCLES));
  }
  else
  {
    MOV(16, R(EAX), Imm16(blockSize[start_addr]));
  }
  JMP(returnDispatcher, true);
}

const u8* DSPEmitter::CompileStub()
{
  const u8* entryPoint = AlignCode16();
  ABI_CallFunction(CompileCurrent);
  XOR(32, R(EAX), R(EAX));  // Return 0 cycles executed
  JMP(returnDispatcher);
  return entryPoint;
}

void DSPEmitter::CompileDispatcher()
{
  enterDispatcher = AlignCode16();
  // We don't use floating point (high 16 bits).
  BitSet32 registers_used = ABI_ALL_CALLEE_SAVED & BitSet32(0xffff);
  ABI_PushRegistersAndAdjustStack(registers_used, 8);

  const u8* dispatcherLoop = GetCodePtr();

  FixupBranch exceptionExit;
  if (Host::OnThread())
  {
    CMP(8, M(const_cast<bool*>(&g_dsp.external_interrupt_waiting)), Imm8(0));
    exceptionExit = J_CC(CC_NE);
  }

  // Check for DSP halt
  TEST(8, M(&g_dsp.cr), Imm8(CR_HALT));
  FixupBranch _halt = J_CC(CC_NE);

  // Execute block. Cycles executed returned in EAX.
  MOVZX(64, 16, ECX, M(&g_dsp.pc));
  MOV(64, R(RBX), ImmPtr(blocks.data()));
  JMPptr(MComplex(RBX, RCX, SCALE_8, 0));

  returnDispatcher = GetCodePtr();

  // Decrement cyclesLeft
  SUB(16, M(&g_cycles_left), R(EAX));

  J_CC(CC_A, dispatcherLoop);

  // DSP gave up the remaining cycles.
  SetJumpTarget(_halt);
  if (Host::OnThread())
  {
    SetJumpTarget(exceptionExit);
  }
  // MOV(32, M(&cyclesLeft), Imm32(0));
  ABI_PopRegistersAndAdjustStack(registers_used, 8);
  RET();
}

}  // namespace x86
}  // namespace JIT
}  // namespace DSP
